/*
 * Examples from Chapter 14.  
 * Binary and N-ary trees.
 */

#include <stdlib.h>
#include <time.h>
#include <glib.h>

gint
my_tree_traverse(
    gpointer    key,
    gpointer    value,
    gpointer    data
)
{
    /*
     * This version of the traverse function is for printing.
     */
    if ( data == 0 )
    {
        g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, (gchar *)(key));
    
        if ( (int)value  > 250 )
        {
            g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "Stopping with value > 250.");
            return TRUE;
        }

        return FALSE;
    }

    /*
     * This version of the traverse function is for freeing up the tree.
     */
    g_free(key);
    return FALSE;

}

void
print_tree(
    GTree        *tree,
    gchar        *header
)
{
    gchar        *string;

    g_print("\n");
    string = g_strdup_printf("*** %s ***", header);
    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, string);
    g_free( string );

    /*
     * Now print out the tree in order.
     */
    g_tree_traverse( tree, my_tree_traverse, G_IN_ORDER, (gpointer) 0 );
}

gboolean
my_node_traverse(
    GNode        *node,
    gpointer    data
)
{
    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, (gchar *)(node->data));
    
    if ( g_node_depth( node ) > 3 )
    {
        g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "Stopping at depth 3.");
        return TRUE;
    }

    return FALSE;
}

void
print_node(
    GNode        *ntree,
    gchar        *header
)
{
    gchar        *string;

    g_print("\n");
    string = g_strdup_printf("*** %s ***", header);
    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, string);
    g_free( string );

    /*
     * Now print out the tree in order.
     */
    g_node_traverse( ntree, G_IN_ORDER, G_TRAVERSE_ALL, -1, 
        my_node_traverse, (gpointer)0 );
}

gint
compare_tree(
    gconstpointer    key1,
    gconstpointer    key2
)
{
    if ( (gint)key1 < (gint)key2 )
        return ( -1 );
    if ( (gint)key1 > (gint)key2 )
        return ( 1 );
    return ( 0 );
}


int main()
{
    GTree        *tree = NULL;
    GNode        *ntree = NULL;
    GNode        *new_node = NULL;
    GNode        *parent = NULL;
    gchar        *string;
    gint        check;
    gint        i, j;


    /*
     * Initialize random number generator.
     */
    srand((int)(
            (1 + (int)(500.0*rand()/(RAND_MAX+1.0))) * time(NULL)*time(NULL)));
    
    /*==========================================
     * Binary trees.
     */

    /*
     * Allocate a tree.
     */
    tree = g_tree_new( (GCompareFunc) compare_tree );

    /*
     * Build a tree with 100 nodes, each with a random key.
     */
    for ( i=0; i<100; i++ )
    {
        string = g_strdup_printf("key %d", i);
        g_tree_insert( tree, string, 
            (gpointer)(1 + (int)(500.0*rand()/(RAND_MAX+1.0))) ); 
    }
    
    /*
     * Now print out the list in order.
     */
    print_tree( tree, "Original tree" );

    /*
     * Print the number of nodes and the height of the tree.
     */
    string = g_strdup_printf("Number of nodes in tree: %d", g_tree_nnodes(tree));
    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, string);
    g_free( string );
    string = g_strdup_printf("Height of tree: %d", g_tree_height(tree));
    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, string);
    g_free( string );
    
    /*
     * Free up the tree 
     */
    g_tree_traverse( tree, my_tree_traverse, G_IN_ORDER, (gpointer) 1 );
    g_tree_destroy( tree );


    /*==========================================
     * N-ary trees.
     */

    /*
     * Allocate a tree.
     */
    string = g_strdup_printf("node %d", 0);
    ntree = g_node_new( (gpointer) string );

    /*
     * Build a tree with 30 nodes, with each node having 3 children.
     * Randomly move up and down into the tree.
     */
    parent = ntree;
    for ( i=1; i<100; i++ )
    {
        string = g_strdup_printf("node %d", i);
        new_node = g_node_append_data( parent, string );
        for ( j=0; j<3; j++ )
        {
            string = g_strdup_printf("child node %d:%d", i, j);
            g_node_append_data( new_node, string );
        }

        check = 1 + (int)(10.0*rand()/(RAND_MAX+1.0));
        if ( check > 7 ) 
        {
            parent = new_node;
        }
        if ( check < 3 ) 
        {
            parent = ntree;
        }
    }
    
    /*
     * Now print out the list in order.
     */
    print_node( ntree, "Original N-ary tree" );

    /*
     * Print the number of nodes and the height of the tree.
     */
    string = 
        g_strdup_printf("Number of nodes in N-ary tree: %d", 
            g_node_n_nodes(ntree, G_TRAVERSE_ALL));
    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, string);
    g_free( string );
    string = g_strdup_printf("Height of N-ary tree: %d", 
            g_node_max_height(ntree));
    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, string);
    g_free( string );
    
   return 0;
}
