#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <gtk/gtk.h>

GtkWidget* fill_tree(GtkWidget *);
GtkWidget* new_tree(GtkTreeItem *);

/* Flag that determines wether the content of the tree should
   always be generated randomly and dynamically, or wether it
   should be fixed in the state in which it is when the flag
   is toggled. */
gint FIXED_CONTENT = FALSE;
/* Flag that determines if lines should be drawn to connect the
   tree items in the same depth or not. */
gint DRAW_LINES = TRUE;

/* Make sure that all the subtrees have the correct settings
   according to the FIXED_CONTENT and DRAW_LINES flags. */
void sync_tree(GtkTree *tree)
{
    GList *glist_tree_item = NULL;
    GtkWidget *subtree = NULL;
    
    /* Apply the state of this flag to the tree. */
    gtk_tree_set_view_lines(tree, DRAW_LINES);

    /* Go over all the children of the tree and obtain their
       eventual subtree. */
    glist_tree_item = tree->children;
    while(glist_tree_item)
    {
        subtree = GTK_TREE_ITEM_SUBTREE(glist_tree_item->data);
        
        /* If the content should be generated dynamically, make
           sure that there is a subtree present. */
        if(!FIXED_CONTENT && !subtree)
        {
            subtree = new_tree(GTK_TREE_ITEM(
                glist_tree_item->data));
            gtk_tree_item_set_subtree(GTK_TREE_ITEM(
                glist_tree_item->data), subtree);
        }
        
        /* If there is a subtree, apply the correct properties
           and synchronize its contents too. */
        if(subtree)
        {
            sync_tree(GTK_TREE(subtree));
            
            /* If the content is fixed, forbit the expansion of
               items when no children are present by removing
               the subtree. */
            if(FIXED_CONTENT &&
               !g_list_length(GTK_TREE(subtree)->children))
            {
                gtk_tree_item_remove_subtree(GTK_TREE_ITEM(
                    glist_tree_item->data));
            }
        }

        /* Get next tree item. */
        glist_tree_item = g_list_next(glist_tree_item);
    }
}

/* Create the full label of a tree item according to its
   ancestor's and its own id in the containing tree. */
gchar* get_tree_item_label(GtkTree *tree, GtkTreeItem *tree_item)
{
    gint id = 0;
    GString *label_string = NULL;
    GString *temp_string = NULL;
    GtkWidget *parent_item = NULL;
    GtkWidget *parent_tree = NULL;
    gchar *label = NULL;

    /* Get the id of the tree item. */
    id = (gint)gtk_object_get_data(GTK_OBJECT(tree_item),
        "id");

    temp_string = g_string_new("");
    label_string = g_string_new("");
    g_string_sprintf(label_string, "%d", id);

    /* Iterate over the parent trees and prepend the ids
       of the items containing them to the label string that's
       being generated. */
    parent_tree = GTK_WIDGET(tree);
    while(parent_tree)
    {
        parent_item = GTK_TREE(parent_tree)->tree_owner;

        if(parent_item)
        {
            g_string_sprintf(temp_string, "%d,",
                (gint)gtk_object_get_data(GTK_OBJECT(
                parent_item), "id"));
            label_string = g_string_prepend(label_string,
                temp_string->str);
            parent_tree = gtk_object_get_data(GTK_OBJECT(
                parent_item), "parent");
        }
        else
        {
            parent_tree = NULL;
        }
    }
    
    label = label_string->str;
    g_string_free(label_string, FALSE);
    g_string_free(temp_string, TRUE);
    
    return label;
}

void print_status_message(GtkTree *tree, gchar *message)
{
    GtkWidget *statusbar = NULL;

    statusbar = gtk_object_get_data(GTK_OBJECT(
        GTK_TREE_ROOT_TREE(tree)), "statusbar");
    gtk_statusbar_pop(GTK_STATUSBAR(statusbar), 1);
    gtk_statusbar_push(GTK_STATUSBAR(statusbar), 1,
        message);
}

void tree_select_child(GtkTree *tree, GtkWidget *tree_item,
         gpointer user_data)
{
    gchar *id = NULL;
    GString *message = NULL;
    
    /* Simulate the handling of a focus item. Store a pointer
       to the last selected item as object data of the root
       tree. */
    gtk_object_set_data(GTK_OBJECT(GTK_TREE_ROOT_TREE(tree)),
        "focus", tree_item);

    /* Create the correct message to display in the status
       bar. */
    id = get_tree_item_label(GTK_TREE(tree),
        GTK_TREE_ITEM(tree_item));
    message = g_string_new("");
    
    /* Verify if the item is still in the GList of selected
       items. If this is not the case, the item was deselected
       instead of selected. */
    if(g_list_find(GTK_TREE_SELECTION(GTK_TREE_ROOT_TREE(
           tree)), tree_item))
    {
        g_string_sprintf(message, " \"%s\" selected ", id);
    }
    else
    {
        g_string_sprintf(message, " \"%s\" deselected ", id);
    }
    
    print_status_message(tree, message->str);
    g_string_free(message, TRUE);
    g_free(id);
}

void collapse_handler(GtkTreeItem *tree_item,
         GtkTree *parent_tree)
{
    gchar *id = NULL;
    GString *message = NULL;
    GtkWidget *tree = NULL;
    
    /* Create the correct message to display in the status
       bar. */
    id = get_tree_item_label(GTK_TREE(parent_tree),
        GTK_TREE_ITEM(tree_item));
    message = g_string_new("");
    g_string_sprintf(message, " \"%s\" collapsed ", id);
    print_status_message(parent_tree, message->str);
    g_string_free(message, TRUE);
    g_free(id);
    
    /* If the content is dynamically generated, remove all the
       children of the collapsed item and provide it with a new
       empty subtree. */
    if(!FIXED_CONTENT)
    {
        tree = GTK_TREE_ITEM_SUBTREE(tree_item);
    
        gtk_tree_clear_items(GTK_TREE(tree), 0, -1);
        if(GTK_TREE_ITEM_SUBTREE(tree_item))
        {
            gtk_tree_item_remove_subtree(GTK_TREE_ITEM(
                tree_item));
        }
        
        gtk_tree_item_set_subtree(tree_item,
            new_tree(tree_item));
    }
}

void expand_handler(GtkTreeItem *tree_item,
         GtkTree *parent_tree)
{
    gchar *id = NULL;
    GString *message = NULL;
    GtkWidget *tree = NULL;
    
    /* Create the correct message to display in the status
       bar. */
    id = get_tree_item_label(GTK_TREE(parent_tree),
        GTK_TREE_ITEM(tree_item));
    message = g_string_new("");
    g_string_sprintf(message, " \"%s\" expanded ", id);
    print_status_message(parent_tree, message->str);
    g_string_free(message, TRUE);
    g_free(id);
    
    /* If the content is dynamically generated, fill the
       sub-tree of the expanded item with child items if the
       sub-tree is empty. */
    if(!FIXED_CONTENT)
    {
        tree = GTK_TREE_ITEM_SUBTREE(tree_item);
    
        if(!g_list_length(GTK_TREE(tree)->children))
        {
            fill_tree(tree);
        }
    }
}

void button_remove_clicked(GtkButton *button, GtkWidget *tree)
{
    GList *tree_items = NULL;
    GtkWidget *tree_item = NULL;
    GtkTree *root_tree = NULL;
    
    root_tree = GTK_TREE(GTK_TREE_ROOT_TREE(tree));
    
    /* Get the tree item that corresponds to the clicked
       remove button. */
    tree_item =
        gtk_object_get_data(GTK_OBJECT(button), "tree_item");
    
    /* Create the GList of tree items that will be removed,in
       this case it will only contain one. */
    tree_items = g_list_prepend(tree_items, tree_item);
    
    /* Remove the tree item. */
    gtk_tree_remove_items(GTK_TREE(tree), tree_items);
    
    sync_tree(root_tree);
}

void button_fixed_toggled(GtkToggleButton *button,
         GtkWidget *tree)
{
    /* Store the new state of this flag. */
    FIXED_CONTENT = gtk_toggle_button_get_active(button);

    /* Synchronize and redraw the entire tree to reflect the
       modification of the flag. */
    sync_tree(GTK_TREE(tree));
    gtk_widget_draw(GTK_WIDGET(tree), NULL);
}

void button_lines_toggled(GtkToggleButton *button,
         GtkWidget *tree)
{
    /* Store the new state of this flag. */
    DRAW_LINES = gtk_toggle_button_get_active(button);

    /* Synchronize and redraw the entire tree to reflect the
       modification of the flag. */
    sync_tree(GTK_TREE(tree));
    gtk_widget_draw(GTK_WIDGET(tree), NULL);
}

void button_remove_selection_clicked(GtkButton *button,
         GtkWidget *tree)
{
    GList *selected_items = NULL;
    GList *items_to_remove = NULL;
    
    /* Get the selected tree items. */
    selected_items = GTK_TREE_SELECTION(tree);
    
    /* Copy the obtained GList to another one since the
       selection GList is changed when removing items. */
    while(selected_items)
    {
        items_to_remove = g_list_prepend(items_to_remove,
            selected_items->data);
        selected_items = g_list_next(selected_items);
    }
    
    /* If there were items in the selection, remove them.*/
    if(items_to_remove)
    {
        items_to_remove = g_list_reverse(items_to_remove);
    
        gtk_tree_remove_items(GTK_TREE(tree), items_to_remove);
    }
    
    /* Synchronize the tree to provide a new subtree should
       the removal of the last child of an item have caused the
       item's subtree to be removed also. */
    sync_tree(GTK_TREE(GTK_TREE_ROOT_TREE(tree)));
}

/* Find a child item in the hierarchy below the provided
   tree. */
gint find_child(GtkTree *tree, GtkWidget *child)
{
    gint position = -1;
    GList *glist_tree_item = NULL;
    GtkWidget *subtree = NULL;

    /* Check if the child is a direct item of the provided
       tree. Return the position if the child item was
       found. */
    position = gtk_tree_child_position(GTK_TREE(tree), child);
    if(position != -1)
    {
        return position;
    }
    
    /* Search the child item in all the subtrees. */
    glist_tree_item = tree->children;
    while(glist_tree_item)
    {
        subtree = GTK_TREE_ITEM_SUBTREE(glist_tree_item->data);
    
        if(subtree)
        {
            position = find_child(GTK_TREE(subtree), child);
            if(position != -1)
            {
                return position;
            }
        }
        glist_tree_item = g_list_next(glist_tree_item);
    }

    return -1;
}

/* Select all the descendants of a tree. */
void select_all_children(GtkTree *tree)
{
    GList *glist_tree_item = NULL;
    GtkWidget *subtree = NULL;

    /* Process all the items of the tree. */
    glist_tree_item = tree->children;
    while(glist_tree_item)
    {
        subtree = GTK_TREE_ITEM_SUBTREE(glist_tree_item->data);
    
        /* Select the item. */
        gtk_tree_item_select(GTK_TREE_ITEM(
            glist_tree_item->data));

        /* Select all the descendants of the subtree. */
        if(subtree)
        {
            select_all_children(GTK_TREE(subtree));
        }
        
        glist_tree_item = g_list_next(glist_tree_item);
    }
}

/* Deselect all the descendants of a tree. */
void deselect_all_children(GtkTree *tree)
{
    GList *glist_tree_item = NULL;
    GtkWidget *subtree = NULL;

    /* Process all the items of the tree. */
    glist_tree_item = tree->children;
    while(glist_tree_item)
    {
        subtree = GTK_TREE_ITEM_SUBTREE(glist_tree_item->data);
    
        /* Deselect the item. */
        gtk_tree_item_deselect(GTK_TREE_ITEM(
            glist_tree_item->data));
            
        /* Deselect all the descendants of the subtree. */
        if(subtree)
        {
            deselect_all_children(GTK_TREE(subtree));
        }

        glist_tree_item = g_list_next(glist_tree_item);
    }
}

GtkWidget* create_tree_item(GtkTree *tree)
{
    gint id = 0;
    GtkWidget *tree_item = NULL;
    GtkWidget *hbox = NULL;
    GtkWidget *button = NULL;
    GtkWidget *label = NULL;
    GtkStyle *newstyle = NULL;
    gchar *label_string = NULL;

    /* Get the id of the last created child of this tree and
       increment it. Store to the new id again. */
    id = (gint)gtk_object_get_data(GTK_OBJECT(tree),
        "last_id");
    id++;
    gtk_object_set_data(GTK_OBJECT(tree), "last_id",
        (gpointer)id);

    /* Create a new tree item, store its id and parent tree and
       connect the expand and collpase signals handlers. */
    tree_item = gtk_tree_item_new();
    gtk_object_set_data(GTK_OBJECT(tree_item), "parent", tree);
    gtk_object_set_data(GTK_OBJECT(tree_item), "id",
        (gpointer)id);
    gtk_signal_connect(GTK_OBJECT(tree_item), "collapse",
        (GtkSignalFunc)collapse_handler, tree);
    gtk_signal_connect(GTK_OBJECT(tree_item), "expand",
        (GtkSignalFunc)expand_handler, tree);

    hbox = gtk_hbox_new(FALSE, 5);
    gtk_container_add(GTK_CONTAINER(tree_item), hbox);

    button = gtk_button_new_with_label(" Remove ");
    gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, TRUE, 0);
    gtk_signal_connect(GTK_OBJECT(button), "clicked",
        (GtkSignalFunc)button_remove_clicked, tree);
    gtk_object_set_data(GTK_OBJECT(button), "tree_item",
        tree_item);

    /* Make the selected button background color the same as
       the normal button background color. */
    gtk_widget_ensure_style(button);
    newstyle = gtk_style_copy(button->style);
    newstyle->bg[GTK_STATE_SELECTED] =
        newstyle->bg[GTK_STATE_NORMAL];
    gtk_widget_set_style(button, newstyle);
    
    /* Make the selected button label foreground color the same
       as the normal button label foreground color and change
       the font. */
    gtk_widget_ensure_style(GTK_BIN(button)->child);
    newstyle = gtk_style_copy(button->style);
    newstyle->fg[GTK_STATE_SELECTED] =
        newstyle->fg[GTK_STATE_NORMAL];
    gdk_font_unref(newstyle->font);
    newstyle->font = gdk_font_load
        ("-adobe-helvetica-bold-r-*-*-*-100-*-*-*-*-*-*");
    gtk_widget_set_style(GTK_BIN(button)->child, newstyle);
    
    label_string = get_tree_item_label(GTK_TREE(tree),
        GTK_TREE_ITEM(tree_item));
    label = gtk_label_new(label_string);
    gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 0);
    g_free(label_string);
    
    gtk_widget_show_all(tree_item);

    return tree_item;
}

/* Retrieve the focused item from a tree. This is only a
   trick to provide the missing functionality. */
GtkWidget *get_focused_item(GtkTree *tree)
{
    GtkWidget *focus_item = NULL;

    focus_item = gtk_object_get_data(GTK_OBJECT(
        GTK_TREE_ROOT_TREE(tree)), "focus");
    
    return focus_item;
}

/* Get the position of the focused item. This can be used to
   check if the focused item is still existing. */
gint get_focused_item_position(GtkTree *tree)
{
    GtkWidget *focus_item = NULL;
    gint focus = -1;

    focus_item = get_focused_item(tree);
    if(focus_item)
    {
        focus = find_child(GTK_TREE(tree), focus_item);
    }
    else
    {
        focus = -1;
    }

    return focus;
}

void button_add_clicked(GtkButton *button, GtkTree *tree)
{
    GtkWidget *focus_item = NULL;
    gint focus = -1;
    GtkWidget *tree_item = NULL;
    GtkTree *parent_tree = NULL;
    
    /* Get the focused item and its position. */
    focus_item = get_focused_item(tree);
    focus = get_focused_item_position(tree);
    
    /* Get the tree into which the new item will be inserted and
       determine at which position this will happen. */
    if(focus == -1)
    {
        parent_tree = GTK_TREE(GTK_TREE_ROOT_TREE(tree));
        focus = 0;
    }
    else
    {
        parent_tree = gtk_object_get_data(GTK_OBJECT(
            focus_item), "parent");
    }
    
    /* Create the new tree item. */
    tree_item = create_tree_item(parent_tree);
    
    /* Insert the item in the parent tree. */
    gtk_tree_insert(parent_tree, tree_item, focus);
    
    /* If the content is generated dynamically, add an empty
       subtree which will make the item display an expand
       symbol. */
    if(!FIXED_CONTENT)
    {
        gtk_tree_item_set_subtree(GTK_TREE_ITEM(tree_item),
            new_tree(GTK_TREE_ITEM(tree_item)));
    }
}

void button_select_subitems_clicked(GtkButton *button,
         GtkTree *tree)
{
    GtkWidget *focus_item = NULL;
    GtkWidget *subtree = NULL;

    /* Check if the focused item is valid. */
    if(get_focused_item_position(tree) != -1)
    {
        /* Get the focused item and its eventual subtree. */
        focus_item = get_focused_item(tree);
        subtree = GTK_TREE_ITEM_SUBTREE(focus_item);
        /* If the focused item exists and it contains a
           subtree, select all children of this subtree. */
        if(focus_item && subtree)
        {
            select_all_children(GTK_TREE(subtree));
        }
    }
}

void button_deselect_subitems_clicked(GtkButton *button,
         GtkTree *tree)
{
    GtkWidget *focus_item = NULL;
    GtkWidget *subtree = NULL;

    /* Check if the focused item is valid. */
    if(get_focused_item_position(tree) != -1)
    {
        /* Get the focused item and its eventual subtree. */
        focus_item = get_focused_item(tree);
        subtree = GTK_TREE_ITEM_SUBTREE(focus_item);
        /* If the focused item exists and it contains a
           subtree, deselect all children of this subtree. */
        if(focus_item && subtree)
        {
            deselect_all_children(GTK_TREE(subtree));
        }
    }
}

GtkWidget* fill_tree(GtkWidget *tree)
{
    gint counter = 0;
    gint items = 0;
    GtkWidget *tree_item = NULL;

    srand(time(NULL));
    items = 1+(gint)(10.0*rand()/RAND_MAX);

    /* Append a random number of items to the tree's content,
       providing each one with a subtree. */
    for(counter = 0; counter < items; counter++)
    {
        tree_item = create_tree_item(GTK_TREE(tree));
        gtk_tree_append(GTK_TREE(tree), tree_item);
        gtk_tree_item_set_subtree(GTK_TREE_ITEM(tree_item),
            new_tree(GTK_TREE_ITEM(tree_item)));
    }

    return tree;
}

GtkWidget* new_tree(GtkTreeItem *tree_item)
{
    GtkWidget *tree = NULL;
    
    /* Create a new GtkTree. */
    tree = gtk_tree_new();
    /* Use the value of the DRAW_LINES flag to initialize this
       setting for the newly created tree. */
    gtk_tree_set_view_lines(GTK_TREE(tree), DRAW_LINES);
    /* Connect a signal handler to the select_child signals. */
    gtk_signal_connect(GTK_OBJECT(tree), "select_child",
        (GtkSignalFunc)tree_select_child, NULL);

    return tree;
}

int main(int argc, char *argv[])
{
    GtkWidget *window = NULL;
    GtkWidget *vbox = NULL;
    GtkWidget *scrolled_window = NULL;
    GtkWidget *tree = NULL;
    GtkWidget *table = NULL;
    GtkWidget *button_fixed = NULL;
    GtkWidget *button_lines = NULL;
    GtkWidget *button_remove_selection = NULL;
    GtkWidget *button_add = NULL;
    GtkWidget *button_select_subitems = NULL;
    GtkWidget *button_deselect_subitems = NULL;
    GtkWidget *statusbar = NULL;

    gtk_init(&argc, &argv);
    
    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(window),
        "Chapter 16 : example 1");
    gtk_window_set_policy(GTK_WINDOW(window), TRUE, TRUE,
        FALSE);
    gtk_window_set_default_size(GTK_WINDOW(window), -1, 300);
    gtk_container_set_border_width(GTK_CONTAINER(window), 5);
    gtk_signal_connect(GTK_OBJECT(window), "delete_event",
        (GtkSignalFunc)gtk_exit, NULL);
    
    vbox = gtk_vbox_new(FALSE, 5);
    gtk_container_add(GTK_CONTAINER(window), vbox);
        
    scrolled_window = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy
        (GTK_SCROLLED_WINDOW(scrolled_window),
         GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    gtk_box_pack_start(GTK_BOX(vbox), scrolled_window,
        TRUE, TRUE, 0);
        
    tree = new_tree(NULL);
    fill_tree(tree);

    gtk_tree_set_selection_mode(GTK_TREE(tree),
        GTK_SELECTION_MULTIPLE);
    gtk_scrolled_window_add_with_viewport(
        GTK_SCROLLED_WINDOW(scrolled_window), tree);

    table = gtk_table_new(3, 2, TRUE);
    gtk_box_pack_start(GTK_BOX(vbox), table,
        FALSE, TRUE, 0);

    button_fixed =
        gtk_toggle_button_new_with_label(" Fix contents ");
    gtk_signal_connect(GTK_OBJECT(button_fixed),
        "toggled",
        (GtkSignalFunc)button_fixed_toggled, tree);
    gtk_table_attach(GTK_TABLE(table), button_fixed,
        0, 1, 0, 1,
        GTK_EXPAND | GTK_SHRINK | GTK_FILL,
        GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
    
    button_lines =
        gtk_toggle_button_new_with_label(" Draw lines ");
    gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(
        button_lines), 1);
    gtk_signal_connect(GTK_OBJECT(button_lines),
        "toggled",
        (GtkSignalFunc)button_lines_toggled, tree);
    gtk_table_attach(GTK_TABLE(table), button_lines,
        1, 2, 0, 1,
        GTK_EXPAND | GTK_SHRINK | GTK_FILL,
        GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
    
    button_remove_selection =
        gtk_button_new_with_label(" Remove selection ");
    gtk_signal_connect(GTK_OBJECT(button_remove_selection),
        "clicked",
        (GtkSignalFunc)button_remove_selection_clicked, tree);
    gtk_table_attach(GTK_TABLE(table), button_remove_selection,
        0, 1, 1, 2,
        GTK_EXPAND | GTK_SHRINK | GTK_FILL,
        GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
    
    button_add = gtk_button_new_with_label(" Add item ");
    gtk_signal_connect(GTK_OBJECT(button_add), "clicked",
        (GtkSignalFunc)button_add_clicked, tree);
    gtk_table_attach(GTK_TABLE(table), button_add,
        1, 2, 1, 2,
        GTK_EXPAND | GTK_SHRINK | GTK_FILL,
        GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
    
    button_select_subitems =
        gtk_button_new_with_label(" Select children ");
    gtk_signal_connect(GTK_OBJECT(button_select_subitems),
        "clicked",
        (GtkSignalFunc)button_select_subitems_clicked, tree);
    gtk_table_attach(GTK_TABLE(table), button_select_subitems,
        0, 1, 2, 3,
        GTK_EXPAND | GTK_SHRINK | GTK_FILL,
        GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
    
    button_deselect_subitems =
        gtk_button_new_with_label(" Deselect children ");
    gtk_signal_connect(GTK_OBJECT(button_deselect_subitems),
        "clicked",
        (GtkSignalFunc)button_deselect_subitems_clicked, tree);
    gtk_table_attach(GTK_TABLE(table),
        button_deselect_subitems,
        1, 2, 2, 3,
        GTK_EXPAND | GTK_SHRINK | GTK_FILL,
        GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
    
    statusbar = gtk_statusbar_new();
    gtk_box_pack_start(GTK_BOX(vbox), statusbar,
        FALSE, TRUE, 0);
    gtk_object_set_data(GTK_OBJECT(tree), "statusbar",
        statusbar);
    
    gtk_widget_show_all(window);
    
    gtk_main();
    
    return 0;
}
