#include <gtk/gtk.h>

/* Define the info fields for the supported targets. */
typedef enum
{
    STRING,
    LENGTH,
} targetInfo;

/* Create the GtkTargetEntry structures array with the correct
   values for the supported targets. */
static GtkTargetEntry target_entries[] =
{
    {"STRING", 0, STRING},
    {"LENGTH", 0, LENGTH}
};

/* Number of targets that are supported. */
static guint ntargets = sizeof(target_entries) /
    sizeof(target_entries[0]);

void register_targets(GtkWidget *widget)
{
    /* Add the supported targets to the provided widget for the
       PRIMARY selection. */
    gtk_selection_add_targets(widget, GDK_SELECTION_PRIMARY,
        target_entries, ntargets);
}

void become_owner(GtkWidget *widget, guint32 time)
{
    /* Array of pointers that will contain the data for each
       target. */
    gpointer *target_data = NULL;
    
    /* Temporary variables to contain each data that will be
       provided. */
    gchar *string = NULL;
    gint  length = 0;
    
    gint  return_value = 0;
    
    /* Try to get the ownership of the PRIMARY selection. */
    return_value = gtk_selection_owner_set(widget,
        GDK_SELECTION_PRIMARY, time);
        
    if(return_value == TRUE)
    {
        g_print("SUCCESSFULLY GOT OWNERSHIP.\n");
    
        /* If this was succcessfull, allocate the array of pointers for
           the target data. */
        target_data = g_new0(gpointer, ntargets);
        gtk_object_set_data(GTK_OBJECT(widget), "target_data",
            target_data);
    
        /* Define the content the targets data. */
        string = "This is the data of the STRING target.";
        length = strlen(string);
    
        /* Make a copy of the target data and store it at the correct
           location in the array. */
        target_data[STRING] = g_strdup(string);
        target_data[LENGTH] = g_memdup(&length,
            sizeof(length));
    }
    else
    {
        g_print("FAILED TO GET OWNERSHIP.\n");
    }
}

void selection_get_handler(GtkWidget *widget,
     GtkSelectionData *data, guint info, guint time,
     gpointer user_data)
{
    /* Array of pointers that contains the data for each
       target. */
    gpointer *target_data =
        gtk_object_get_data(GTK_OBJECT(widget), "target_data");
    
    gchar *target_name = gdk_atom_name(data->target);
    g_print("Got request for the '%s' target.\n", target_name);
    g_free(target_name);
    
    /* Provide different data according to the target info
       field that has been sent. */
    switch(info)
    {
        case STRING :
            gtk_selection_data_set(data,
                gdk_atom_intern("STRING", TRUE), /* type   */
                8,                               /* format */
                target_data[STRING],             /* data   */
                strlen(target_data[STRING]));    /* length */
            break;
        case LENGTH :
            gtk_selection_data_set(data,
                gdk_atom_intern("INTEGER", TRUE), /* type   */
                32,                               /* format */
                target_data[LENGTH],              /* data   */
                sizeof(target_data[LENGTH]));     /* length */
            break;
    }
}

gboolean selection_clear_handler(GtkWidget *widget,
         GdkEventSelection *event, gpointer user_data)
{
    /* Array of pointers that contains the data for each
       target. */
    gpointer *target_data =
        gtk_object_get_data(GTK_OBJECT(widget), "target_data");
    
    g_print("FORCEFULLY CLEARED OWNERSHIP.\n");
    
    /* Release the memory that was allocated when the ownership
       was obtained. */
    g_free(target_data[STRING]);
    g_free(target_data[LENGTH]);
    g_free(target_data);
    
    gtk_object_remove_data(GTK_OBJECT(widget), "target_data");
    
    return FALSE;
}

void button_button_release(GtkWidget *button,
     GdkEventButton *event, gpointer user_data)
{
    become_owner(button, event->time);
}

int main(int argc, char *argv[])
{
    GtkWidget *window = NULL;
    GtkWidget *button = NULL;
    
    gtk_init(&argc, &argv);
    
    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(window),
        "Chapter 20 : example 2");
    gtk_signal_connect(GTK_OBJECT(window), "delete_event",
        (GtkSignalFunc)gtk_exit, NULL);
    
    button = gtk_button_new_with_label(" Become OWNER ");
    gtk_container_add(GTK_CONTAINER(window), button);
    gtk_signal_connect(GTK_OBJECT(button),
        "button_release_event",
        (GtkSignalFunc)button_button_release, NULL);
    gtk_signal_connect(GTK_OBJECT(button),
        "selection_get",
        (GtkSignalFunc)selection_get_handler, NULL);
    gtk_signal_connect(GTK_OBJECT(button),
        "selection_clear_event",
        (GtkSignalFunc)selection_clear_handler, NULL);
    
    register_targets(button);
    
    gtk_widget_show_all(window);
    
    gtk_main();
    
    return 0;
}
