#include <time.h>
#include <gtk/gtk.h>
#include "sams.xpm"

/* Define the info fields for the supported targets. */
typedef enum
{
    TEXT_HTML,
    STRING
} targetInfo;

/* Create the GtkTargetEntry structures array with the correct
   values for the supported targets. */
static GtkTargetEntry target_entries[] =
{
    {"text/html", GTK_TARGET_SAME_APP, TEXT_HTML},
    {"STRING",    0,                   STRING}
};

/* Number of targets that are supported. */
static guint ntargets = sizeof(target_entries) /
    sizeof(target_entries[0]);

void drag_data_received_handler(GtkWidget *widget,
         GdkDragContext *drag_context, gint x, gint y,
         GtkSelectionData *data, guint info, guint event_time,
         gpointer user_data)
{
    time_t    now = 0;
    struct tm *now_tm = NULL;
    gchar     now_string[10];
    gchar     *label = NULL;
    gchar     *target_name = NULL;
    
    now = time(NULL);
    now_tm = localtime(&now);
    strftime(now_string, sizeof(now_string), "%T", now_tm);
    label = g_strconcat(" Drag DESTINATION received data at ",
        now_string, " ", NULL);
    gtk_label_set(GTK_LABEL(GTK_BIN(widget)->child), label);
    g_free(label);
    
    target_name = gdk_atom_name(data->target);
    g_print("D: Received data for the '%s' target.\n",
        target_name);
    g_free(target_name);
    
    switch(info)
    {
        case TEXT_HTML :
            g_print("text/html = '%s'\n", (gchar*)data->data);
            break;
        case STRING :
            g_print("string = '%s'\n", (gchar*)data->data);
            break;
    }
}

void drag_begin_handler(GtkWidget *widget,
         GdkDragContext *drag_context, gpointer user_data)
{
    /* Array of pointers that will contain the data for each
       target. */
    gpointer *target_data = NULL;
    
    /* Temporary pointers to each data that will be
       provided. */
    gchar    *html = NULL;
    gchar    *string = NULL;
    time_t   now = time(NULL);
    gchar    *now_string = ctime(&now);
    
    g_print("S: Drag begun\n");
    
    /* Allocate the array of pointers for the target data. */
    target_data = g_new0(gpointer, ntargets);
    gtk_object_set_data(GTK_OBJECT(widget), "target_data",
        target_data);
    
    /* Define the content the targets data. */
    html = g_strconcat("<P>It's now <B>", now_string,
               "</B>.</P>", NULL);
    string = g_strconcat("It's now ", now_string, ".", NULL);
    
    /* Make a copy of the target data and store it at the
       correct location in the array. */
    target_data[TEXT_HTML] = g_strdup(html);
    target_data[STRING] = g_strdup(string);
}

void drag_data_get_handler(GtkWidget *widget,
         GdkDragContext *drag_context, GtkSelectionData *data,
         guint info, guint time, gpointer user_data)
{
    /* Array of pointers that contains the data for each
       target. */
    gpointer *target_data =
        gtk_object_get_data(GTK_OBJECT(widget), "target_data");
    
    gchar *target_name = gdk_atom_name(data->target);
    g_print("S: Got request for the '%s' target.\n", target_name);
    g_free(target_name);
    
    switch(info)
    {
        case TEXT_HTML :
            gtk_selection_data_set(data,
                gdk_atom_intern("text/html", TRUE),/* type   */
                8,                                 /* format */
                target_data[TEXT_HTML],            /* data   */
                strlen(target_data[TEXT_HTML]));   /* length */
            break;
        case STRING :
            gtk_selection_data_set(data,
                gdk_atom_intern("STRING", TRUE), /* type   */
                8,                               /* format */
                target_data[STRING],             /* data   */
                strlen(target_data[STRING]));    /* length */
            break;
    }
}

void drag_end_handler(GtkWidget *widget,
         GdkDragContext *drag_context, gpointer user_data)
{
    /* Array of pointers that contains the data for each
       target. */
    gpointer *target_data =
        gtk_object_get_data(GTK_OBJECT(widget), "target_data");
    
    g_print("S: Drag ended\n");
    
    /* Release the memory that was allocated when widget was
       set as a drag source. */
    g_free(target_data[TEXT_HTML]);
    g_free(target_data[STRING]);
    g_free(target_data);
    
    gtk_object_remove_data(GTK_OBJECT(widget), "target_data");
}

void make_drag_destination(GtkWidget *widget)
{
    gtk_drag_dest_set(widget, GTK_DEST_DEFAULT_ALL,
        target_entries, ntargets, GDK_ACTION_COPY);
}

void make_drag_source(GtkWidget *widget)
{
    GdkPixmap *drag_icon;
    GdkBitmap *drag_mask;
    
    gtk_drag_source_set(widget, GDK_BUTTON1_MASK,
        target_entries, ntargets, GDK_ACTION_COPY);
    
    drag_icon = gdk_pixmap_colormap_create_from_xpm_d(NULL,
	                gdk_colormap_get_system(), &drag_mask,
                    NULL, sams_xpm);
    
    gtk_drag_source_set_icon(widget, gdk_colormap_get_system(),
        drag_icon, drag_mask);
    
    gdk_pixmap_unref(drag_icon);
    gdk_pixmap_unref(drag_mask);
}

int main(int argc, char *argv[])
{
    GtkWidget *window = NULL;
    GtkWidget *vbox = NULL;
    GtkWidget *button_destination = NULL;
    GtkWidget *button_source = NULL;
    
    gtk_init(&argc, &argv);
    
    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(window),
        "Chapter 21 : example 1");
    gtk_signal_connect(GTK_OBJECT(window), "delete_event",
        (GtkSignalFunc)gtk_exit, NULL);
        
    vbox = gtk_vbox_new(TRUE, 5);
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
    gtk_container_add(GTK_CONTAINER(window), vbox);
    
    button_destination =
        gtk_button_new_with_label(" Drag DESTINATION received"
            " nothing");
    gtk_signal_connect(GTK_OBJECT(button_destination),
        "drag_data_received",
        (GtkSignalFunc)drag_data_received_handler, NULL);
    gtk_box_pack_start_defaults(GTK_BOX(vbox),
        button_destination);
    
    button_source =
        gtk_button_new_with_label(" Drag SOURCE providing"
            " current time ");
    gtk_signal_connect(GTK_OBJECT(button_source),
        "drag_begin",
        (GtkSignalFunc)drag_begin_handler, NULL);
    gtk_signal_connect(GTK_OBJECT(button_source),
        "drag_data_get",
        (GtkSignalFunc)drag_data_get_handler, NULL);
    gtk_signal_connect(GTK_OBJECT(button_source),
        "drag_end",
        (GtkSignalFunc)drag_end_handler, NULL);
    gtk_box_pack_end_defaults(GTK_BOX(vbox), button_source);
    
    make_drag_destination(button_destination);
    make_drag_source(button_source);
    
    gtk_widget_show_all(window);
    
    gtk_main();
    
    return 0;
}
