#include <time.h>
#include <gdk/gdktypes.h>
#include <gtk/gtk.h>

/* Define the info fields for the supported targets. */
typedef enum
{
    TEXT_HTML = 1,
    STRING = 2
} targetInfo;

/* Create the GtkTargetEntry structures array with the correct
   values for the supported targets. */
static GtkTargetEntry target_entries[] =
{
    {"text/html", 0, TEXT_HTML},
    {"STRING",    0, STRING}
};

/* Number of targets that are supported. */
static guint ntargets = sizeof(target_entries) /
    sizeof(target_entries[0]);

void drag_data_received_handler(GtkWidget *widget,
         GdkDragContext *drag_context, gint x, gint y,
         GtkSelectionData *data, guint info, guint event_time,
         gpointer user_data)
{
    time_t    now = 0;
    struct tm *now_tm = NULL;
    gchar     now_string[10];
    gchar     *label = NULL;
    gchar     *target_name = NULL;
    
    now = time(NULL);
    now_tm = localtime(&now);
    strftime(now_string, sizeof(now_string), "%T", now_tm);
    label = g_strconcat(" Drag DESTINATION received data at ",
        now_string, " ", NULL);
    gtk_label_set(GTK_LABEL(GTK_BIN(widget)->child), label);
    g_free(label);
    
    target_name = gdk_atom_name(data->target);
    g_print("D: Received data for the '%s' target.\n",
        target_name);
    
    switch(info)
    {
        case TEXT_HTML :
            g_print("'%s' = '%s'\n", target_name, (gchar*)data->data);
            gtk_drag_finish(drag_context, TRUE, FALSE, event_time);
            break;
        case STRING :
            g_print("'%s' = '%s'\n", target_name, (gchar*)data->data);
            gtk_drag_finish(drag_context, TRUE, FALSE, event_time);
            break;
        default :
            gtk_drag_finish(drag_context, FALSE, FALSE, event_time);
            break;
    }

    g_free(target_name);
}

GdkAtom find_common_atom(GdkDragContext *drag_context)
{
    gint    counter = 0;
    GdkAtom target_atom = GDK_NONE;

    /* Check if there are common targets between the drag
       source and the drag destination. */
    for(counter = 0; counter < ntargets; counter++)
    {
        target_atom =
            gdk_atom_intern(target_entries[counter].target,
                FALSE);

        /* When the first common target is found, return it. */
        if(g_list_find(drag_context->targets,
           (gpointer)target_atom))
        {
            return target_atom;
        }
    }
    
    return GDK_NONE;
}

gboolean drag_motion_handler(GtkWidget *widget,
             GdkDragContext *drag_context, gint x, gint y,
             guint time, gpointer user_data)
{
    /* If a common target was found, notify the drag source
       that the destination widget is ready to accept a drop by
       using the action that was suggested by Gtk. If this
       suggestion is 0, the destination isn't ready to accept a
       drop. */
    if(find_common_atom(drag_context) != GDK_NONE)
    {
        gdk_drag_status(drag_context,
            drag_context->suggested_action, time);
            
        /* If the destination widget isn't highlighted yet
           and an action has been suggested by Gtk,
           higlight the destination widget. */
        if(!gtk_object_get_data(GTK_OBJECT(widget),
           "highlighted") &&
           drag_context->suggested_action)
        {
            g_print("D: Entered destination widget.\n");
            gtk_object_set_data(GTK_OBJECT(widget),
                "highlighted", (gpointer)TRUE);
            gtk_drag_highlight(widget);
        }
        
        /* The widget is a valid destination. */
        return TRUE;
    }

    /* If there are no common targets, the destination widget
       will not accept a drop. */
    gdk_drag_status(drag_context, 0, time);

    /* The widget is a valid destination. */
    return TRUE;
}

void drag_leave_handler(GtkWidget *widget,
         GdkDragContext *drag_context, guint time,
         gpointer user_data)
{
    g_print("D: Left destination widget.\n");
    gtk_drag_unhighlight(widget);
    gtk_object_remove_data(GTK_OBJECT(widget), "highlighted");
}

gboolean drag_drop_handler(GtkWidget *widget,
             GdkDragContext *drag_context, gint x, gint y,
             guint time, gpointer user_data)
{
    GdkAtom target_atom = GDK_NONE;
    
    g_print("D: Dropping !\n");
    
    target_atom = find_common_atom(drag_context);
    
    /* When the a common target is found, its data is requested
       from the source widget. */
    if(target_atom != GDK_NONE)
    {            
        gtk_drag_get_data(widget, drag_context,
            target_atom, time);
    }
    
    return TRUE;
}

void draw_drag_icon(GtkWidget *widget,
         GdkDragContext *drag_context, gchar *text)
{
    GtkStyle  *style = NULL;
    gint      width = 0;
    gint      height = 0;
    GdkPixmap *icon_pixmap = NULL;
    GdkPixmap *icon_mask = NULL;
    GdkGC      *pixmap_gc = NULL;
    GdkGC      *mask_gc = NULL;
    GdkColor  mask_color;
    
    /* Get the GtkStyle of the drag source widget. */
    gtk_widget_ensure_style(widget);
    style = gtk_widget_get_style(widget);
        
    /* Calculate the geometry of the drag icon. */
    width = gdk_string_width(style->font, text) + 4;
    height = style->font->ascent + style->font->descent;

    /* Allocate the icon pixmap and mask*/
    icon_pixmap = gdk_pixmap_new(widget->window, width, height,
                      -1);
    icon_mask = gdk_pixmap_new(NULL, width, height, 1);
    
    /* Initialize the icon pixmap and mask graphical
       contexts */
    pixmap_gc = gdk_gc_new(icon_pixmap);
    gdk_gc_set_background(pixmap_gc, &style->white);
    gdk_gc_set_foreground(pixmap_gc, &style->black);
    gdk_gc_set_function(pixmap_gc, GDK_COPY);
    gdk_gc_set_font(pixmap_gc, style->font);
    
    mask_gc = gdk_gc_new(icon_mask);
    mask_color.pixel = 1;
    gdk_gc_set_foreground(mask_gc, &mask_color);
    gdk_gc_set_function(mask_gc, GDK_COPY);

    /* Set the drag icon background */
    gdk_gc_set_foreground(pixmap_gc, &style->white);
    gdk_draw_rectangle(icon_pixmap, pixmap_gc, TRUE, 0, 0,
        width, height);
    gdk_gc_set_foreground(pixmap_gc, &style->black);

    /* Make the icon completely opaque. */
    gdk_draw_rectangle(icon_mask, mask_gc, TRUE, 0, 0,
        width, height);
    mask_color.pixel = 0;
    gdk_gc_set_foreground(mask_gc, &mask_color);

    /* Draw the text in the icon pixmap. */
    gdk_draw_text(icon_pixmap, style->font, pixmap_gc, 2,
        style->font->ascent, text, strlen(text));

    /* Set the new icon. */
    gtk_drag_set_icon_pixmap(drag_context,
        gdk_colormap_get_system(), icon_pixmap, icon_mask,
        width/2, height/2);
}

void drag_begin_handler(GtkWidget *widget,
         GdkDragContext *drag_context, gpointer user_data)
{
    /* Array of pointers that will contain the data for each
       target. */
    gpointer *target_data = NULL;
    
    /* Temporary pointers to each data format that will be
       provided. */
    gchar  *html = NULL;
    gchar  *string = NULL;
    time_t now = time(NULL);
    gchar  *now_string = ctime(&now);
    
    g_print("S: Drag begun\n");

    /* Draw a drag icon that contains the current time. */
    draw_drag_icon(widget, drag_context, now_string);
    
    /* Allocate the array of pointers for the target data. */
    target_data = g_new0(gpointer, ntargets);
    gtk_object_set_data(GTK_OBJECT(widget), "target_data",
        target_data);
    
    /* Define the content the targets data. */
    html = g_strconcat("<P>It's now <B>", now_string,
               "</B>.</P>", NULL);
    string = g_strconcat("It's now ", now_string, ".", NULL);
    
    /* Make a copy of the target data and store it at the
       correct location in the array. */
    target_data[TEXT_HTML] = g_strdup(html);
    target_data[STRING] = g_strdup(string);
}

void drag_data_get_handler(GtkWidget *widget,
         GdkDragContext *drag_context, GtkSelectionData *data,
         guint info, guint time, gpointer user_data)
{
    /* Array of pointers that contains the data for each
       target. */
    gpointer *target_data =
        gtk_object_get_data(GTK_OBJECT(widget), "target_data");
    
    gchar *target_name = gdk_atom_name(data->target);
    g_print("S: Got request for the '%s' target.\n", target_name);
    g_free(target_name);
    
    switch(info)
    {
        case TEXT_HTML :
            gtk_selection_data_set(data,
                gdk_atom_intern("text/html", TRUE),/* type   */
                8,                                 /* format */
                target_data[TEXT_HTML],            /* data   */
                strlen(target_data[TEXT_HTML]));   /* length */
            break;
        case STRING :
            gtk_selection_data_set(data,
                gdk_atom_intern("STRING", TRUE), /* type   */
                8,                               /* format */
                target_data[STRING],             /* data   */
                strlen(target_data[STRING]));    /* length */
            break;
    }
}

void drag_end_handler(GtkWidget *widget,
         GdkDragContext *drag_context, gpointer user_data)
{
    /* Array of pointers that contains the data for each
       target. */
    gpointer *target_data =
        gtk_object_get_data(GTK_OBJECT(widget), "target_data");
    
    g_print("S: Drag ended\n");
    
    /* Release the memory that was allocated when widget was
       set as a drag source. */
    g_free(target_data[TEXT_HTML]);
    g_free(target_data[STRING]);
    g_free(target_data);
    
    gtk_object_remove_data(GTK_OBJECT(widget), "target_data");
}

gboolean button_press_handler(GtkWidget *widget,
             GdkEventButton *event, gpointer user_data)
{
    /* Check if the left mouse button was pressed. */
    if(event->button == 1)
    {
        /* Check if the control key was pressed. */
        if(event->state == GDK_CONTROL_MASK)
        {
            /* Define what type of drag operation will be
               executed. */
            gtk_object_set_data(GTK_OBJECT(widget),
                "drag_type", (gpointer)STRING);
        }
        /* Check if the control and alt keys were pressed. */
        else if(event->state == (GDK_CONTROL_MASK |
                                 GDK_MOD1_MASK))
        {
            /* Define what type of drag operation will be
               executed. */
            gtk_object_set_data(GTK_OBJECT(widget),
                "drag_type", (gpointer)TEXT_HTML);
        }
    }
    
    return TRUE;
}

gboolean button_release_handler(GtkWidget *widget,
             GdkEventButton *event, gpointer user_data)
{
    gtk_object_remove_data(GTK_OBJECT(widget), "drag_type");
    
    return TRUE;
}

gboolean motion_notify_handler(GtkWidget *widget,
             GdkEventMotion *event, gpointer user_data)
{
    gint          drag_type = 0;
    GtkTargetList *target_list = NULL;
    
    /* Get the type of drag operation that will be executed. */
    drag_type = (gint)gtk_object_get_data(GTK_OBJECT(widget),
                    "drag_type");
    
    /* According to each type, provide different targets, if a
       drag & drop operation has been initiated. */
    if(drag_type == STRING)
    {
        target_list = gtk_target_list_new(target_entries,
                          ntargets);
        /* This doesn't work with gtk 1.2.6 and before due to a
           bug. A patch has been submitted and it should be
           fixed in later versions. */
        gtk_drag_begin(widget, target_list, GDK_ACTION_COPY, 1,
            (GdkEvent*)event);
        gtk_target_list_unref(target_list);
    }
    else if(drag_type == TEXT_HTML)
    {
        target_list = gtk_target_list_new(target_entries,
                          ntargets);
        /* This doesn't work with gtk 1.2.6 and before due to a
           bug. A patch has been submitted and it should be
           fixed in later versions. */
        gtk_target_list_remove(target_list,
            gdk_atom_intern("STRING", FALSE));
        gtk_drag_begin(widget, target_list, GDK_ACTION_COPY, 1,
            (GdkEvent*)event);
        gtk_target_list_unref(target_list);
    }    
    
    return TRUE;
}

int main(int argc, char *argv[])
{
    GtkWidget *window = NULL;
    GtkWidget *vbox = NULL;
    GtkWidget *button_destination = NULL;
    GtkWidget *button_source = NULL;
    
    gtk_init(&argc, &argv);
    
    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(window),
        "Chapter 21 : example 2");
    gtk_signal_connect(GTK_OBJECT(window), "delete_event",
        (GtkSignalFunc)gtk_exit, NULL);
        
    vbox = gtk_vbox_new(TRUE, 5);
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
    gtk_container_add(GTK_CONTAINER(window), vbox);
    
    button_destination =
        gtk_button_new_with_label(" Drag DESTINATION received"
            " nothing");
    gtk_drag_dest_set(button_destination,
        0, target_entries, ntargets,
        GDK_ACTION_COPY);
    gtk_signal_connect(GTK_OBJECT(button_destination),
        "drag_data_received",
        (GtkSignalFunc)drag_data_received_handler, NULL);
    gtk_signal_connect(GTK_OBJECT(button_destination),
        "drag_motion",
        (GtkSignalFunc)drag_motion_handler, NULL);
    gtk_signal_connect(GTK_OBJECT(button_destination),
        "drag_leave",
        (GtkSignalFunc)drag_leave_handler, NULL);
    gtk_signal_connect(GTK_OBJECT(button_destination),
        "drag_drop",
        (GtkSignalFunc)drag_drop_handler, NULL);
    gtk_box_pack_start_defaults(GTK_BOX(vbox),
        button_destination);
    
    button_source =
        gtk_button_new_with_label(" Drag SOURCE providing"
            " current time ");
    gtk_widget_add_events(button_source,
        gtk_widget_get_events(button_source) |
        GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
        GDK_BUTTON_MOTION_MASK);
    gtk_signal_connect(GTK_OBJECT(button_source),
        "button_press_event",
        (GtkSignalFunc)button_press_handler, NULL);
    gtk_signal_connect(GTK_OBJECT(button_source),
        "button_release_event",
        (GtkSignalFunc)button_release_handler, NULL);
    gtk_signal_connect(GTK_OBJECT(button_source),
        "motion_notify_event",
        (GtkSignalFunc)motion_notify_handler, NULL);
    gtk_signal_connect(GTK_OBJECT(button_source),
        "drag_begin",
        (GtkSignalFunc)drag_begin_handler, NULL);
    gtk_signal_connect(GTK_OBJECT(button_source),
        "drag_data_get",
        (GtkSignalFunc)drag_data_get_handler, NULL);
    gtk_signal_connect(GTK_OBJECT(button_source),
        "drag_end",
        (GtkSignalFunc)drag_end_handler, NULL);
    gtk_box_pack_end_defaults(GTK_BOX(vbox), button_source);
    
    gtk_widget_show_all(window);
    
    gtk_main();
    
    return 0;
}
